# -*- coding: utf-8 -*-
"""Utilites for Cognite API SDK

This module provides helper methods and different utilities for the Cognite API Python SDK.

This module is protected and should not used by end-users.
"""
import gzip
import json
import re
import sys
import time
from datetime import datetime

import requests

import cognite.config as config


def get_request(url, params=None, headers=None, cookies=None):
    '''Perform a GET request with a predetermined number of retries.'''
    for number_of_tries in range(config.get_number_of_retries() + 1):
        try:
            res = requests.get(url, params=params, headers=headers, cookies=cookies)
            if res.status_code == 200:
                return res
        except Exception as e:
            if number_of_tries == config.get_number_of_retries():
                raise APIError(e)
    try:
        err_mess = res.json()['error'].__str__()
    except:
        err_mess = res.content.__str__()
    err_mess += '\nX-Request_id: {}'.format(res.headers.get('X-Request-Id'))
    raise APIError(err_mess)


def post_request(url, body, headers=None, params=None, cookies=None, use_gzip=False):
    '''Perform a POST request with a predetermined number of retries.'''
    for number_of_tries in range(config.get_number_of_retries() + 1):
        try:
            if use_gzip:
                if headers:
                    headers['Content-Encoding'] = 'gzip'
                else:
                    headers = {'Content-Encoding': 'gzip'}
                res = requests.post(url,
                                    data=gzip.compress(json.dumps(body).encode('utf-8')),
                                    headers=headers,
                                    params=params,
                                    cookies=cookies
                                    )
            else:
                res = requests.post(url,
                                    data=json.dumps(body),
                                    headers=headers,
                                    params=params,
                                    cookies=cookies
                                    )
            if res.status_code == 200:
                return res
        except Exception as e:
            if number_of_tries == config.get_number_of_retries():
                raise APIError(e)
    try:
        err_mess = res.json()['error'].__str__()
    except:
        err_mess = res.content.__str__()
    err_mess += '\nX-Request_id: {}'.format(res.headers.get('X-Request-Id'))
    raise APIError(err_mess)


def put_request(url, body=None, headers=None, cookies=None):
    '''Perform a PUT request with a predetermined number of retries.'''
    for number_of_tries in range(config.get_number_of_retries() + 1):
        try:
            res = requests.put(url, data=json.dumps(body), headers=headers, cookies=cookies)
            if res.ok:
                return res
        except Exception as e:
            if number_of_tries == config.get_number_of_retries():
                raise APIError(e)
    try:
        err_mess = res.json()['error'].__str__()
    except:
        err_mess = res.content.__str__()
    err_mess += '\nX-Request_id: {}'.format(res.headers.get('X-Request-Id'))
    raise APIError(err_mess)


def datetime_to_ms(dt):
    epoch = datetime.utcfromtimestamp(0)
    return int((dt - epoch).total_seconds() * 1000)


def round_to_nearest(x, base):
    return int(base * round(float(x) / base))


def granularity_to_ms(time_string):
    '''Returns millisecond representation of granularity time string'''
    magnitude = int(''.join([c for c in time_string if c.isdigit()]))
    unit = ''.join([c for c in time_string if c.isalpha()])
    unit_in_ms = {'s': 1000, 'second': 1000, 'm': 60000, 'minute': 60000,
                  'h': 3600000, 'hour': 3600000, 'd': 86400000, 'day': 86400000}
    return magnitude * unit_in_ms[unit]


def _time_ago_to_ms(time_ago_string):
    '''Returns millisecond representation of time-ago string'''
    pattern = r'(\d+)([a-z])-ago'
    res = re.match(pattern, str(time_ago_string))
    if res:
        magnitude = int(res.group(1))
        unit = res.group(2)
        unit_in_ms = {'s': 1000, 'm': 60000, 'h': 3600000, 'd': 86400000, 'w': 604800000}
        return magnitude * unit_in_ms[unit]
    return None


def interval_to_ms(start, end):
    '''Returns the ms representation of start-end-interval whether it is time-ago, datetime or None.'''
    time_now = int(round(time.time() * 1000))
    if isinstance(start, datetime):
        start = datetime_to_ms(start)
    elif isinstance(start, str):
        start = time_now - _time_ago_to_ms(start)
    elif start is None:
        start = time_now - _time_ago_to_ms('1w-ago')

    if isinstance(end, datetime):
        end = datetime_to_ms(end)
    elif isinstance(end, str):
        end = time_now - _time_ago_to_ms(end)
    elif end is None:
        end = time_now

    return start, end


class APIError(Exception):
    pass


class ProgressIndicator():
    '''This class lets the system give the user that data is being downloaded'''

    def __init__(self, tag_ids):
        sys.stdout.write("\rDownloading data for tags {}...".format(tag_ids))

    def terminate(self):
        sys.stdout.write('\r' + ' ' * 500 + '\r')



# -*- coding: utf-8 -*-
"""Raw Module

This module mirrors the Raw API. It allows the user to handle raw data.

https://doc.cognitedata.com/#Cognite-API-Cloud-Raw
"""

from typing import List

import cognite._utils as _utils
import cognite.config as config
from cognite.data_objects import RawObject, RawRowDTO


def get_databases(
        limit: int = None,
        cursor: str = None,
        api_key=None,
        project=None
):
    """Returns a RawObject containing a list of raw databases.

    Args:
        limit (int):    A limit on the amount of results to return.

        cursor (str):   A cursor can be provided to navigate through pages of results.

        api_key (str):  Your api-key.

        project (str):  Project name.

    Returns:
        RawObject: A data object containing the requested data with several getter methods with different
        output formats.
    """
    api_key, project = config.get_config_variables(api_key, project)
    url = config.get_base_url() + '/projects/{}/raw'.format(project)

    params = {
        'limit': limit,
        'cursor': cursor
    }

    headers = {
        'api-key': api_key,
        'content-type': '*/*',
        'accept': 'application/json'
    }

    res = _utils.get_request(url=url, params=params, headers=headers, cookies=config.get_cookies())
    return RawObject(res.json())


def create_databases(
        database_names: list,
        api_key=None,
        project=None
):
    """Creates databases in the Raw API and returns the created databases.

    Args:
        database_names (list):  A list of databases to create.

        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        RawObject: A data object containing the requested data with several getter methods with different
        output formats.

    """
    api_key, project = config.get_config_variables(api_key, project)
    url = config.get_base_url() + '/projects/{}/raw/create'.format(project)
    body = {
        'items': [{'dbName': '{}'.format(database_name)} for database_name in database_names]
    }
    headers = {
        'api-key': api_key,
        'content-type': '*/*',
        'accept': 'application/json'
    }
    res = _utils.post_request(url=url, body=body, headers=headers, cookies=config.get_cookies())
    return RawObject(res.json())


def delete_databases(
        database_names: list,
        recursive: bool = False,
        api_key=None,
        project=None
):
    """Deletes databases in the Raw API.

    Args:
        database_names (list):  A list of databases to delete.

        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        An empty response.

    """
    api_key, project = config.get_config_variables(api_key, project)
    url = config.get_base_url() + '/projects/{}/raw/delete'.format(project)
    body = {
        'items': [{'dbName': '{}'.format(database_name)} for database_name in database_names]
    }
    params = {
        'recursive': recursive
    }
    headers = {
        'api-key': api_key,
        'content-type': '*/*',
        'accept': 'application/json'
    }
    res = _utils.post_request(url=url, body=body, params=params, headers=headers, cookies=config.get_cookies())
    return res.json()


def get_tables(
        database_name: str = None,
        limit: int = None,
        cursor: str = None,
        api_key=None,
        project=None
):
    """Returns a RawObject containing a list of tables in a raw database.

    Args:
        database_name (str):   The database name to retrieve tables from.

        limit (int):    A limit on the amount of results to return.

        cursor (str):   A cursor can be provided to navigate through pages of results.

        api_key (str):  Your api-key.

        project (str):  Project name.

    Returns:
        RawObject: A data object containing the requested data with several getter methods with different
        output formats.
    """
    api_key, project = config.get_config_variables(api_key, project)
    url = config.get_base_url() + '/projects/{}/raw/{}'.format(project, database_name)
    params = dict()
    if not limit:
        params['limit'] = limit
    if not cursor:
        params['cursor'] = cursor
    headers = {
        'api-key': api_key,
        'content-type': '*/*',
        'accept': 'application/json'
    }
    res = _utils.get_request(url=url, params=params, headers=headers, cookies=config.get_cookies())
    return RawObject(res.json())


def create_tables(
        database_name: str = None,
        table_names: list = None,
        api_key=None,
        project=None
):
    """Creates tables in the given Raw API database.

    Args:
        database_name (str):    The database to create tables in.

        table_names (list):     The table names to create.

        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        RawObject: A data object containing the requested data with several getter methods with different
        output formats.

    """
    api_key, project = config.get_config_variables(api_key, project)
    url = config.get_base_url() + '/projects/{}/raw/{}/create'.format(project, database_name)
    body = {
        'items': [{'tableName': '{}'.format(table_name)} for table_name in table_names]
    }
    headers = {
        'api-key': api_key,
        'content-type': '*/*',
        'accept': 'application/json'
    }
    res = _utils.post_request(url=url, body=body, headers=headers, cookies=config.get_cookies())
    return RawObject(res.json())


def delete_tables(
        database_name: str = None,
        table_names: list = None,
        api_key=None,
        project=None
):
    """Deletes databases in the Raw API.

    Args:
        database_name (str):    The database to create tables in.

        table_names (list):     The table names to create.

        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        An empty response.

    """
    api_key, project = config.get_config_variables(api_key, project)
    url = config.get_base_url() + '/projects/{}/raw/{}/delete'.format(project, database_name)
    body = {
        'items': [{'tableName': '{}'.format(table_name)} for table_name in table_names]
    }
    headers = {
        'api-key': api_key,
        'content-type': '*/*',
        'accept': 'application/json'
    }
    res = _utils.post_request(url=url, body=body, headers=headers, cookies=config.get_cookies())
    return res.json()


def get_rows(
        database_name: str = None,
        table_name: str = None,
        limit: int = None,
        cursor: str = None,
        api_key=None,
        project=None
):
    """Returns a RawObject containing a list of rows.

    Args:
        database_name (str):    The database name to retrieve rows from.

        table_name (str):       The table name to retrieve rows from.

        limit (int):            A limit on the amount of results to return.

        cursor (str):           A cursor can be provided to navigate through pages of results.

        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        RawObject: A data object containing the requested data with several getter methods with different
        output formats.
    """
    api_key, project = config.get_config_variables(api_key, project)
    url = config.get_base_url() + '/projects/{}/raw/{}/{}'.format(project, database_name, table_name)
    params = dict()
    params['limit'] = limit
    params['cursor'] = cursor
    headers = {
        'api-key': api_key,
        'content-type': '*/*',
        'accept': 'application/json'
    }
    res = _utils.get_request(url=url, params=params, headers=headers, cookies=config.get_cookies())
    return RawObject(res.json())


def create_rows(
        database_name: str = None,
        table_name: str = None,
        rows: List[RawRowDTO] = None,
        api_key=None,
        project=None,
        ensure_parent=False,
        use_gzip=False
):
    """Creates tables in the given Raw API database.

    Args:
        database_name (str):    The database to create rows in.

        table_name (str):       The table names to create rows in.

        rows (list):            The rows to create.

        api_key (str):          Your api-key.

        project (str):          Project name.

        ensure_parent (bool):   Create database/table if it doesn't exist already

        use_gzip (bool):        Compress content using gzip

    Returns:
        An empty response

    """
    api_key, project = config.get_config_variables(api_key, project)
    url = config.get_base_url() + '/projects/{}/raw/{}/{}/create'.format(project, database_name, table_name)

    headers = {
        'api-key': api_key,
        'content-type': '*/*',
        'accept': 'application/json'
    }
    if ensure_parent:
        params = {'ensureParent': 'true'}
    else:
        params = {}

    ul_row_limit = 1000
    i = 0
    while i < len(rows):
        body = {'items': [{'key': '{}'.format(row.key), 'columns': row.columns} for row in rows[i:i + ul_row_limit]]}
        res = _utils.post_request(url=url, body=body, headers=headers, params=params, cookies=config.get_cookies(),
                                  use_gzip=use_gzip)
        i += ul_row_limit
    return res.json()


def delete_rows(
        database_name: str = None,
        table_name: str = None,
        rows: List[RawRowDTO] = None,
        api_key=None,
        project=None
):
    """Deletes rows in the Raw API.

    Args:
        database_name (str):    The database to create tables in.

        table_name (str):      The table name where the rows are at.

        rows (list):            The rows to delete.

        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        An empty response.

    """
    api_key, project = config.get_config_variables(api_key, project)
    url = config.get_base_url() + '/projects/{}/raw/{}/{}/delete'.format(project, database_name, table_name)
    body = {
        'items': [
            {
                'key': '{}'.format(row.key),
                'columns': row.columns
            }
            for row in rows
        ]
    }
    headers = {
        'api-key': api_key,
        'content-type': '*/*',
        'accept': 'application/json'
    }
    res = _utils.post_request(url=url, body=body, headers=headers, cookies=config.get_cookies())
    return res.json()


def get_row(
        database_name: str = None,
        table_name: str = None,
        row_key: str = None,
        api_key=None,
        project=None
):
    """Returns a RawObject containing a list of rows.

    Args:
        database_name (str):    The database name to retrieve rows from.

        table_name (str):       The table name to retrieve rows from.

        row_key (str):          The key of the row to fetch.

        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        RawObject: A data object containing the requested data with several getter methods with different
        output formats.
    """
    api_key, project = config.get_config_variables(api_key, project)
    url = config.get_base_url() + '/projects/{}/raw/{}/{}/{}'.format(project, database_name, table_name, row_key)
    params = dict()
    headers = {
        'api-key': api_key,
        'content-type': '*/*',
        'accept': 'application/json'
    }
    res = _utils.get_request(url=url, params=params, headers=headers, cookies=config.get_cookies())
    return RawObject(res.json())



# -*- coding: utf-8 -*-
"""Cloud Storage Module

This module mirrors the Cloud Storage API. It allows you to manage files in cloud storage.

https://doc.cognitedata.com/#Cognite-API-Cloud-Storage
"""

import os

import requests

import cognite._utils as _utils
import cognite.config as config


def upload_file(file_name, file_path=None, directory=None, source=None, file_type=None, content_type=None, **kwargs):
    '''Upload metadata about a file and get an upload link.

    The link will expire after 30 seconds if not resumable. A resumable upload link is default. Such a link is one-time
    use and expires after one week. For more information, check this link:
    https://cloud.google.com/storage/docs/json_api/v1/how-tos/resumable-upload.Use PUT request to upload file with the
    link returned.

    Args:
        file_name (str):      File name. Max length is 256.

        file_path (str, optional):     Path to file to upload, if omitted a upload link will be returned.

        content_type (str, optional):   MIME type of your file. Required if file_path is specified.

        directory (str, optional):      Directory containing the file. Max length is 512.

        source (str, optional):         Source that this file comes from. Max length is 256.

        file_type (str, optional):      File type. E.g. pdf, css, spreadsheet, .. Max length is 64.

    Keyword Args:
        api_key (str, optional):        Your api-key.

        project (str, optional):        Project name.

        metadata (dict):      Customized data about the file.

        tag_ids (list):       IDs (tagIds) of equipment related to this file.

        resumable (bool):     Whether to generate a resumable URL or not. Default is true.

        overwrite (bool):     Whether to overwrite existing data if duplicate or not. Default is false.

    Returns:
        dict: A dictionary containing the field fileId and optionally also uploadURL if file_path is omitted.
    '''
    api_key, project = config.get_config_variables(kwargs.get('api_key'), kwargs.get('project'))
    url = config.get_base_url() + '/projects/{}/storage/metadata'.format(project)

    headers = {
        'api-key': api_key,
        'content-type': 'application/json',
        'accept': 'application/json',
        'X-Upload-Content-Type': content_type
    }

    params = {
        'resumable': kwargs.get('resumable', True),
        'overwrite': kwargs.get('overwrite', False)
    }

    body = {
        'fileName': file_name,
        'directory': directory,
        'source': source,
        'fileType': file_type,
        'metadata': kwargs.get('metadata', None),
        'tagIds': kwargs.get('tagIds', None)
    }
    res_storage = _utils.post_request(url=url, body=body, headers=headers, params=params, cookies=config.get_cookies())
    result = res_storage.json()['data']
    if file_path:
        if not content_type:
            return {'message': 'content-type must be specified'}
        headers = {
            'content-length': str(os.path.getsize(file_path))
        }
        with open(file_path, 'rb') as file:
            requests.put(result['uploadURL'], data=file, headers=headers)
        result.pop('uploadURL')
    return result


def list_files(id=None, name=None, directory=None, file_type=None, source=None, **kwargs):
    '''Get list of files matching query.

    This method will automate paging for the user and return info about all files for the given query. If limit is
    specified the method will not page and return that many results.

    Args:
        name (str, optional):      List all files with this name.

        id (int, optional):             List only the file with this id.

        directory (str, optional):      Directory to list files from.

        source (str, optional):         List files coming from this source.

        file_type (str, optional):      Type of files to list.

    Keyword Args:
        api_key (str, optional):        Your api-key.

        project (str, optional):        Project name.

        tag_id (list):                  Returns all files associated with this tagId.

        sort (str):                     Sort descending or ascending. 'ASC' or 'DESC'.

        limit (int):                    Number of results to return.

    Returns:
        list: A list of files with file info.
    '''
    api_key, project = config.get_config_variables(kwargs.get('api_key'), kwargs.get('project'))
    url = config.get_base_url() + '/projects/{}/storage'.format(project)
    headers = {
        'api-key': api_key,
        'accept': 'application/json'
    }
    params = {
        'tagId': kwargs.get('tag_id', None),
        'dir': directory,
        'name': name,
        'type': file_type,
        'source': source,
        'isUploaded': kwargs.get('is_uploaded', None),
        'sort': kwargs.get('sort', None),
        'limit': kwargs.get('limit', 100)
    }
    file_list = []
    res = _utils.get_request(url=url, headers=headers, params=params, cookies=config.get_cookies())
    file_list.extend(res.json()['data']['items'])
    next_cursor = res.json()['data'].get('nextCursor', None)
    limit = kwargs.get('limit', float('inf'))
    while next_cursor is not None and len(file_list) < limit:
        params['cursor'] = next_cursor
        res = _utils.get_request(url=url, headers=headers, params=params, cookies=config.get_cookies())
        file_list.extend(res.json()['data']['items'])
        next_cursor = res.json()['data'].get('nextCursor', None)
    if id is not None:
        return [file for file in file_list if file['id'] == int(id)]
    return file_list


def download_file(id, get_contents=False, **kwargs):
    '''Get list of files matching query.

    Args:
        id (int):                           Path to file to upload, if omitted a upload link will be returned.

        get_contents (bool, optional):      Boolean to determince whether or not to return file contents as string.
                                            Default is False and download url is returned.

    Keyword Args:
        api_key (str, optional):            Your api-key.

        project (str, optional):            Project name.

    Returns:
        str: Download link if get_contents is False else file contents.
    '''
    api_key, project = config.get_config_variables(kwargs.get('api_key'), kwargs.get('project'))
    url = config.get_base_url() + '/projects/{}/storage/{}'.format(project, id)
    headers = {
        'api-key': api_key,
        'accept': 'application/json'
    }
    res = _utils.get_request(url=url, headers=headers, cookies=config.get_cookies())
    if get_contents:
        dl_link = res.json()['data']
        res = requests.get(dl_link)
        return res.content
    return res.json()['data']



# -*- coding: utf-8 -*-
"""Tag Matching Module

This module mirrors the Tag Matching API. It allows the user to search for tag id matches.

https://doc.cognitedata.com/#Cognite-API-Tag-Matching
"""
import cognite._utils as _utils
import cognite.config as config
from cognite.data_objects import TagMatchingObject


def tag_matching(tag_ids, fuzzy_threshold=0, platform=None, **kwargs):
    '''Returns a TagMatchingObject containing a list of matched tags for the given query.

    This method takes an arbitrary string as argument and performs fuzzy matching with a user defined threshold
    toward tag ids in the system.

    Args:
        tag_ids (list):         The tag_ids to retrieve matches for.

        fuzzy_threshold (int):  The threshold to use when searching for matches. A fuzzy threshold of 0 means you only
                                want to accept perfect matches. Must be >= 0.

        platform (str):         The platform to search on.

    Keyword Args:
        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        TagMatchingObject: A data object containing the requested data with several getter methods with different
        output formats.
    '''
    api_key, project = config.get_config_variables(kwargs.get('api_key'), kwargs.get('project'))
    url = config.get_base_url() + '/projects/{}/tagmatching'.format(project)
    body = {
        'tagIds': tag_ids,
        'metadata': {
            'fuzzyThreshold': fuzzy_threshold,
            'platform': platform
        }
    }
    headers = {
        'api-key': api_key,
        'content-type': '*/*',
        'accept': 'application/json'
    }
    res = _utils.post_request(url=url, body=body, headers=headers, cookies=config.get_cookies())
    return TagMatchingObject(res.json())



# -*- coding: utf-8 -*-
"""Timeseries Module

This module mirrors the Timeseries API. It allows you to fetch data from the api and output it in various formats.

https://doc.cognitedata.com/#Cognite-API-Time-series
"""
import io
import os
import warnings
from datetime import datetime
from functools import partial
from multiprocessing import Pool
from typing import List

import pandas as pd

import cognite._constants as _constants
import cognite._utils as _utils
import cognite.config as config
from cognite._protobuf_descriptors import _api_timeseries_data_v1_pb2
from cognite.data_objects import DatapointsObject, LatestDatapointObject, DatapointDTO, TimeSeriesDTO, TimeseriesObject


def get_datapoints(tag_id, aggregates=None, granularity=None, start=None, end=None, **kwargs):
    '''Returns a DatapointsObject containing a list of datapoints for the given query.

    This method will automate paging for the user and return all data for the given time period.

    Args:
        tag_id (str):           The tag_id to retrieve data for.

        aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions
                                are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.

        granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
                                second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.

        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since
                                    epoch or a datetime object which will be converted to ms since epoch UTC.

        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.

    Keyword Arguments:
        protobuf (bool):        Download the data using the binary protobuf format. Only applicable when getting raw data.
                                Defaults to True.

        processes (int):        Number of download processes to run in parallell. Defaults to number returned by cpu_count().

        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        DatapointsObject: A data object containing the requested data with several getter methods with different
        output formats.
    '''
    api_key, project = config.get_config_variables(kwargs.get('api_key'), kwargs.get('project'))
    start, end = _utils.interval_to_ms(start, end)

    diff = end - start
    num_of_processes = kwargs.get('processes', os.cpu_count())

    granularity_ms = 1
    if granularity:
        granularity_ms = _utils.granularity_to_ms(granularity)

    # Ensure that number of steps is not greater than the number data points that will be returned
    steps = min(num_of_processes, max(1, int(diff / granularity_ms)))
    # Make step size a multiple of the granularity requested in order to ensure evenly spaced results
    step_size = _utils.round_to_nearest(int(diff / steps), base=granularity_ms)
    # Create list of where each of the parallelized intervals will begin
    step_starts = [start + (i * step_size) for i in range(steps)]
    args = [{'start': start, 'end': start + step_size} for start in step_starts]

    partial_get_dps = partial(
        _get_datapoints_helper_wrapper,
        tag_id=tag_id,
        aggregates=aggregates,
        granularity=granularity,
        protobuf=kwargs.get('protobuf', True),
        api_key=api_key,
        project=project
    )

    prog_ind = _utils.ProgressIndicator([tag_id])

    p = Pool(steps)

    datapoints = p.map(partial_get_dps, args)
    p.close()
    p.join()
    concat_dps = []
    [concat_dps.extend(el) for el in datapoints]

    prog_ind.terminate()

    return DatapointsObject({'data': {'items': [{'tagId': tag_id, 'datapoints': concat_dps}]}})


def _get_datapoints_helper_wrapper(args, tag_id, aggregates, granularity, protobuf, api_key, project):
    return _get_datapoints_helper(
        tag_id,
        aggregates,
        granularity,
        args['start'],
        args['end'],
        protobuf=protobuf,
        api_key=api_key,
        project=project,
    )


def _get_datapoints_helper(tag_id, aggregates=None, granularity=None, start=None, end=None, **kwargs):
    '''Returns a list of datapoints for the given query.

    This method will automate paging for the given time period.

    Args:
        tag_id (str):           The tag_id to retrieve data for.

        aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions
                                are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.

        granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
                                second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.

        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since
                                    epoch or a datetime object which will be converted to ms since epoch UTC.

        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.

    Keyword Arguments:
        protobuf (bool):        Download the data using the binary protobuf format. Only applicable when getting raw data.
                                Defaults to True.

        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        list of datapoints: A list containing datapoint dicts.
    '''
    api_key, project = kwargs.get('api_key'), kwargs.get('project')
    tag_id = tag_id.replace('/', '%2F')
    url = config.get_base_url() + '/projects/{}/timeseries/data/{}'.format(project, tag_id)

    use_protobuf = kwargs.get('protobuf', True) and aggregates is None
    limit = _constants.LIMIT if aggregates is None else _constants.LIMIT_AGG

    params = {
        'aggregates': aggregates,
        'granularity': granularity,
        'limit': limit,
        'start': start,
        'end': end,
    }

    headers = {
        'api-key': api_key,
        'accept': 'application/protobuf' if use_protobuf else 'application/json'
    }
    datapoints = []
    while not datapoints or len(datapoints[-1]) == limit:
        res = _utils.get_request(url, params=params, headers=headers)
        if use_protobuf:
            ts_data = _api_timeseries_data_v1_pb2.TimeseriesData()
            ts_data.ParseFromString(res.content)
            res = [{'timestamp': p.timestamp, 'value': p.value} for p in ts_data.numericData.points]
        else:
            res = res.json()['data']['items'][0]['datapoints']

        if not res:
            warning = "An interval with no data has been requested ({}, {})." \
                .format(params['start'], params['end'])
            warnings.warn(warning)
            break

        datapoints.append(res)
        latest_timestamp = int(datapoints[-1][-1]['timestamp'])
        params['start'] = latest_timestamp + (_utils.granularity_to_ms(granularity) if granularity else 1)
    dps = []
    [dps.extend(el) for el in datapoints]
    return dps


def post_datapoints(tag_id, datapoints: List[DatapointDTO], **kwargs):
    '''Insert a list of datapoints.

    Args:
        tag_id (str):       ID of timeseries to insert to.

        datapoints (list[DatapointDTO): List of datapoint data transfer objects to insert.

    Keyword Args:
        api_key (str): Your api-key.

        project (str): Project name.

    Returns:
        An empty response.
    '''
    api_key, project = config.get_config_variables(kwargs.get('api_key'), kwargs.get('project'))
    tag_id = tag_id.replace('/', '%2F')
    url = config.get_base_url() + '/projects/{}/timeseries/data/{}'.format(project, tag_id)

    headers = {
        'api-key': api_key,
        'content-type': 'application/json',
        'accept': 'application/json'
    }

    ul_dps_limit = 100000
    i = 0
    while i < len(datapoints):
        body = {'items': [dp.__dict__ for dp in datapoints[i:i + ul_dps_limit]]}
        res = _utils.post_request(url, body=body, headers=headers)
        i += ul_dps_limit
        print(i)
    return res.json()


def get_latest(tag_id, **kwargs):
    '''Returns a LatestDatapointObject containing the latest datapoint for the given tag_id.

    Args:
        tag_id (str):           The tag_id to retrieve data for.

    Keyword Arguments:
        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        DatapointsObject: A data object containing the requested data with several getter methods with different
        output formats.
    '''
    api_key, project = config.get_config_variables(kwargs.get('api_key'), kwargs.get('project'))
    tag_id = tag_id.replace('/', '%2F')
    url = config.get_base_url() + '/projects/{}/timeseries/latest/{}'.format(project, tag_id)
    headers = {
        'api-key': api_key,
        'accept': 'application/json'
    }
    res = _utils.get_request(url, headers=headers, cookies=config.get_cookies())
    return LatestDatapointObject(res.json())


def get_multi_tag_datapoints(tag_ids, aggregates=None, granularity=None, start=None, end=None, limit=None, **kwargs):
    '''Returns a list of DatapointsObjects each of which contains a list of datapoints for the given tag_id.

    Args:
        tag_ids (list):                 The list of tag_ids to retrieve data for.

        aggregates (list, optional):    The list of aggregate functions you wish to apply to the data. Valid aggregate
                                        functions are: 'average/avg, max, min, count, sum, interpolation/int,
                                        stepinterpolation/step'.

        granularity (str):              The granularity of the aggregate values. Valid entries are : 'day/d, hour/h,
                                        minute/m, second/s', or a multiple of these indicated by a number as a prefix
                                        e.g. '12hour'.

        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since
                                    epoch or a datetime object which will be converted to ms since epoch UTC.

        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.

        limit (int):                    Return up to this number of datapoints.

    Keyword Arguments:
        api_key (str):                  Your api-key.

        project (str):                  Project name.

    Returns:
        list(DatapointsObject): A list of data objects containing the requested data with several getter methods
        with different output formats.
    '''
    api_key, project = config.get_config_variables(kwargs.get('api_key'), kwargs.get('project'))
    url = config.get_base_url() + '/projects/{}/timeseries/dataquery'.format(project)

    if isinstance(start, datetime):
        start = _utils.datetime_to_ms(start)
    if isinstance(end, datetime):
        end = _utils.datetime_to_ms(end)

    if limit is None:
        limit = _constants.LIMIT if aggregates is not None else _constants.LIMIT_AGG

    body = {
        'items': [{'tagId': '{}'.format(tag_id)}
                  if isinstance(tag_id, str)
                  else {'tagId': '{}'.format(tag_id['tagId']), 'aggregates': tag_id.get('aggregates', [])} for tag_id in
                  tag_ids],
        'aggregates': aggregates,
        'granularity': granularity,
        'start': start,
        'end': end,
        'limit': limit
    }
    headers = {
        'api-key': api_key,
        'content-type': 'application/json',
        'accept': 'application/json'
    }
    res = _utils.post_request(url=url, body=body, headers=headers, cookies=config.get_cookies())
    return [DatapointsObject({'data': {'items': [dp]}}) for dp in res.json()['data']['items']]


def get_datapoints_frame(tag_ids, aggregates, granularity, start=None, end=None, **kwargs):
    '''Returns a pandas dataframe of datapoints for the given tag_ids all on the same timestamps.

    This method will automate paging for the user and return all data for the given time period.

    Args:
        tag_ids (list):     The list of tag_ids to retrieve data for. Each tag_id can be either a string containing the
                            tag_id or a dictionary containing the tag_id and a list of specific aggregate functions.

        aggregates (list):  The list of aggregate functions you wish to apply to the data for which you have not
                            specified an aggregate function. Valid aggregate functions are: 'average/avg, max, min,
                            count, sum, interpolation/int, stepinterpolation/step'.

        granularity (str):  The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
                            second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.

        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since
                                    epoch or a datetime object which will be converted to ms since epoch UTC.

        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.

    Keyword Arguments:
        api_key (str): Your api-key.

        project (str): Project name.

        processes (int):    Number of download processes to run in parallell. Defaults to number returned by cpu_count().

    Returns:
        pandas.DataFrame: A pandas dataframe containing the datapoints for the given tag_ids. The datapoints for all the
        tag_ids will all be on the same timestamps.

    Note:
        The ``tag_ids`` parameter can take a list of strings and/or dicts on the following formats::

            Using strings:
                ['<tag_id1>', '<tag_id2>']

            Using dicts:
                [{'tagId': '<tag_id1>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']},
                {'tagId': '<tag_id2>', 'aggregates': []}]

            Using both:
                ['<tagid1>', {'tagId': '<tag_id2>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']}]
    '''
    api_key, project = config.get_config_variables(kwargs.get('api_key'), kwargs.get('project'))

    start, end = _utils.interval_to_ms(start, end)

    diff = end - start
    num_of_processes = kwargs.get('processes', os.cpu_count())

    granularity_ms = 1
    if granularity:
        granularity_ms = _utils.granularity_to_ms(granularity)

    # Ensure that number of steps is not greater than the number data points that will be returned
    steps = min(num_of_processes, max(1, int(diff / granularity_ms)))
    # Make step size a multiple of the granularity requested in order to ensure evenly spaced results
    step_size = _utils.round_to_nearest(int(diff / steps), base=granularity_ms)
    # Create list of where each of the parallelized intervals will begin
    step_starts = [start + (i * step_size) for i in range(steps)]
    args = [{'start': start, 'end': start + step_size} for start in step_starts]

    partial_get_dpsf = partial(
        _get_datapoints_frame_helper_wrapper,
        tag_ids=tag_ids,
        aggregates=aggregates,
        granularity=granularity,
        api_key=api_key,
        project=project
    )

    prog_ind = _utils.ProgressIndicator(tag_ids)
    p = Pool(steps)

    dataframes = p.map(partial_get_dpsf, args)
    p.close()
    p.join()
    df = pd.concat(dataframes).drop_duplicates(subset='timestamp').reset_index(drop=True)

    prog_ind.terminate()

    return df


def _get_datapoints_frame_helper_wrapper(args, tag_ids, aggregates, granularity, api_key, project):
    return _get_datapoints_frame_helper(
        tag_ids,
        aggregates,
        granularity,
        args['start'],
        args['end'],
        api_key=api_key,
        project=project
    )


def _get_datapoints_frame_helper(tag_ids, aggregates, granularity, start=None, end=None, **kwargs):
    '''Returns a pandas dataframe of datapoints for the given tag_ids all on the same timestamps.

    This method will automate paging for the user and return all data for the given time period.

    Args:
        tag_ids (list):     The list of tag_ids to retrieve data for. Each tag_id can be either a string containing the
                            tag_id or a dictionary containing the tag_id and a list of specific aggregate functions.

        aggregates (list):  The list of aggregate functions you wish to apply to the data for which you have not
                            specified an aggregate function. Valid aggregate functions are: 'average/avg, max, min,
                            count, sum, interpolation/int, stepinterpolation/step'.

        granularity (str):  The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
                            second/s', or a multiple of these indicated by a number as a prefix e.g. '12hour'.

        start (Union[str, int, datetime]):    Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.
                                    E.g. '2d-ago' will get everything that is up to 2 days old. Can also send time in ms since
                                    epoch or a datetime object which will be converted to ms since epoch UTC.

        end (Union[str, int, datetime]):      Get datapoints up to this time. Same format as for start.

    Keyword Arguments:
        api_key (str): Your api-key.

        project (str): Project name.

    Returns:
        pandas.DataFrame: A pandas dataframe containing the datapoints for the given tag_ids. The datapoints for all the
        tag_ids will all be on the same timestamps.

    Note:
        The ``tag_ids`` parameter can take a list of strings and/or dicts on the following formats::

            Using strings:
                ['<tag_id1>', '<tag_id2>']

            Using dicts:
                [{'tagId': '<tag_id1>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']},
                {'tagId': '<tag_id2>', 'aggregates': []}]

            Using both:
                ['<tagid1>', {'tagId': '<tag_id2>', 'aggregates': ['<aggfunc1>', '<aggfunc2>']}]
    '''
    api_key, project = kwargs.get('api_key'), kwargs.get('project')
    url = config.get_base_url() + '/projects/{}/timeseries/dataframe'.format(project)

    num_aggregates = 0
    num_agg_per_tag = len(aggregates)
    for tag_id in tag_ids:
        if isinstance(tag_id, str):
            num_aggregates += num_agg_per_tag
        else:
            num_aggregates += len(tag_id['aggregates'])
    per_tag_limit = int(_constants.LIMIT / num_aggregates)
    body = {
        'items': [{'tagId': '{}'.format(tag_id)}
                  if isinstance(tag_id, str)
                  else {'tagId': '{}'.format(tag_id['tagId']), 'aggregates': tag_id.get('aggregates', [])} for tag_id in
                  tag_ids],
        'aggregates': aggregates,
        'granularity': granularity,
        'start': start,
        'end': end,
        'limit': per_tag_limit
    }
    headers = {
        'api-key': api_key,
        'content-type': 'application/json',
        'accept': 'text/csv'
    }
    dataframes = []
    while not dataframes or dataframes[-1].shape[0] == per_tag_limit:
        res = _utils.post_request(url=url, body=body, headers=headers, cookies=config.get_cookies())
        dataframes.append(
            pd.read_csv(io.StringIO(res.content.decode(res.encoding if res.encoding else res.apparent_encoding))))
        if dataframes[-1].empty:
            warning = "An interval with no data has been requested ({}, {})." \
                .format(body['start'], body['end'])
            warnings.warn(warning)
            break
        latest_timestamp = int(dataframes[-1].iloc[-1, 0])
        body['start'] = latest_timestamp + _utils.granularity_to_ms(granularity)
    return pd.concat(dataframes).reset_index(drop=True)


def get_timeseries(prefix=None, description=None, include_metadata=False, asset_id=None, path=None, **kwargs):
    '''Returns a TimeseriesObject containing the requested timeseries.

    This method will automate paging for the user and return info about all files for the given query. If limit is
    specified the method will not page and return that many results.

    Args:
        prefix (str):           List timeseries with this prefix in the name.

        description (str):      Filter timeseries taht contains this string in its description.

        include_metadata (bool):    Decide if teh metadata field should be returned or not. Defaults to False.

        asset_id (int):        Get timeseries related to this asset.

        path (str):             Get timeseries under this asset path branch.

    Keyword Arguments:
        limit (int):            Number of results to return.

        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        TimeseriesObject: A data object containing the requested timeseries with several getter methods with different
        output formats.
    '''
    api_key, project = config.get_config_variables(kwargs.get('api_key'), kwargs.get('project'))
    url = config.get_base_url() + '/projects/{}/timeseries'.format(project)
    headers = {
        'api-key': api_key,
        'accept': 'application/json'
    }
    params = {
        'q': prefix,
        'description': description,
        'includeMetadata': include_metadata,
        'assetId': asset_id,
        'path': path,
        'limit': kwargs.get('limit', 10000)
    }
    timeseries = []

    res = _utils.get_request(url=url, headers=headers, params=params, cookies=config.get_cookies())
    timeseries.extend(res.json()['data']['items'])
    next_cursor = res.json()['data'].get('nextCursor', None)
    limit = kwargs.get('limit', float('inf'))

    while next_cursor is not None and len(timeseries) < limit:
        params['cursor'] = next_cursor
        res = _utils.get_request(url=url, headers=headers, params=params, cookies=config.get_cookies())
        timeseries.extend(res.json()['data']['items'])
        next_cursor = res.json()['data'].get('nextCursor', None)

    return TimeseriesObject(timeseries)


def post_time_series(time_series: List[TimeSeriesDTO], **kwargs):
    '''Create a new time series.

    Args:
        timeseries (list[TimeSeriesDTO]):   List of time series data transfer objects to create.

    Keyword Args:
        api_key (str): Your api-key.

        project (str): Project name.
    Returns:
        An empty response.
    '''

    api_key, project = config.get_config_variables(kwargs.get('api_key'), kwargs.get('project'))
    url = config.get_base_url() + '/projects/{}/timeseries'.format(project)

    body = {
        'items': [ts.__dict__ for ts in time_series]
    }

    headers = {
        'api-key': api_key,
        'content-type': 'application/json',
        'accept': 'application/json'
    }

    res = _utils.post_request(url, body=body, headers=headers)
    return res.json()


def update_time_series(time_series: List[TimeSeriesDTO], **kwargs):
    '''Update an existing time series.

    For each field that can be updated, a null value indicates that nothing should be done.

    Args:
        timeseries (list[TimeSeriesDTO]):   List of time series data transfer objects to create.

    Keyword Args:
        api_key (str): Your api-key.

        project (str): Project name.

    Returns:
        An empty response.
    '''

    api_key, project = config.get_config_variables(kwargs.get('api_key'), kwargs.get('project'))
    url = config.get_base_url() + '/projects/{}/timeseries'.format(project)

    body = {
        'items': [ts.__dict__ for ts in time_series]
    }

    headers = {
        'api-key': api_key,
        'content-type': 'application/json',
        'accept': 'application/json'
    }

    res = _utils.put_request(url, body=body, headers=headers)
    return res.json()



# -*- coding: utf-8 -*-
"""Assets Module.

This module mirrors the Assets API.

https://doc.cognitedata.com/#Cognite-API-Assets
"""
from typing import List

import cognite._constants as constants
import cognite._utils as utils
import cognite.config as config
from cognite.data_objects import AssetSearchObject, AssetDTO


# Author: TK
def get_assets(name=None, path=None, description=None, metadata=None, depth=None, fuzziness=None, **kwargs):
    '''Returns assets matching provided description.

    Args:
        name (str):             The name of the asset(s) to get.

        path (str):             The path of the subtree to search in.

        description (str):      Search query.

        metadata (str):         The metadata values used to filter the results.

        depth (str):            Get sub assets up oto this many levels below the specified path.

        fuzziness (int):        The degree of fuzziness in the name matching.

    Keyword Arguments:
        limit (int):            The maximum number of assets to be returned.

        cursor (str):           Cursor to use for paging through results.

        api_key (str):          Your api-key.

        project (str):          Project name.
    Returns:
        AssetSearchObject: A data object containing the requested assets with several getter methods with different
        output formats.
    '''
    api_key, project = config.get_config_variables(kwargs.get('api_key'), kwargs.get('project'))
    url = config.get_base_url() + '/projects/{}/assets'.format(project)
    params = {
        'name': name,
        'description': description,
        'path': path,
        'metadata': metadata,
        'depth': depth,
        'fuzziness': fuzziness,
        'cursor': kwargs.get('cursor'),
        'limit': kwargs.get('limit', constants.LIMIT)
    }
    headers = {
        'api-key': api_key,
        'accept': 'application/json'
    }
    res = utils.get_request(url, params=params, headers=headers, cookies=config.get_cookies())
    return AssetSearchObject(res.json())


# Author: TK
def get_asset_subtree(asset_id='', depth=None, **kwargs):
    '''Returns assets with provided assetId.

    Args:
        asset_id (str):         The asset id of the top asset to get.

        depth (int):            Get subassets this many levels below the top asset.

    Keyword Arguments:
        limit (int):            The maximum nuber of assets to be returned.

        cursor (str):           Cursor to use for paging through results.

        api_key (str):          Your api-key.

        project (str):          Project name.
    Returns:
        AssetSearchObject: A data object containing the requested assets with several getter methods with different
        output formats.
    '''
    api_key, project = config.get_config_variables(kwargs.get('api_key'), kwargs.get('project'))
    url = config.get_base_url() + '/projects/{}/assets/{}'.format(project, asset_id)
    params = {
        'depth': depth,
        'limit': kwargs.get('limit', constants.LIMIT),
        'cursor': kwargs.get('cursor')
    }
    headers = {
        'api-key': api_key,
        'accept': 'application/json'
    }
    res = utils.get_request(url, params=params, headers=headers, cookies=config.get_cookies())
    return AssetSearchObject(res.json())


def post_assets(assets: List[AssetDTO], **kwargs):
    '''Insert a list of assets.

    Args:
        assets (list[AssetDTO]): List of asset data transfer objects.

    Keyword Args:
        api_key (str): Your api-key.

        project (str): Project name.

    Returns:
        AssetSearchObject: A data object containing the posted assets with several getter methods with different
        output formats.
    '''
    api_key, project = config.get_config_variables(kwargs.get('api_key'), kwargs.get('project'))
    url = config.get_base_url() + '/projects/{}/assets'.format(project)

    body = {
        'items': [asset.__dict__ for asset in assets]
    }

    headers = {
        'api-key': api_key,
        'content-type': 'application/json',
        'accept': 'application/json'
    }

    res = utils.post_request(url, body=body, headers=headers, cookies=config.get_cookies())
    return AssetSearchObject(res.json())



# -*- coding: utf-8 -*-
'''Project Configuration Module.

This module allows you to set an api-key and a project for your python project.
'''
from cognite._constants import BASE_URL, RETRY_LIMIT

_CONFIG_API_KEY = ''
_CONFIG_PROJECT = ''
_CONFIG_COOKIES = {}
_CONFIG_BASE_URL = None
_CONFIG_RETRIES = None


def configure_session(api_key='', project='', cookies=None):
    '''Sets session variables.

    Args:
        api_key (str):  Api-key for current project.

        cookies (dict): Cookies to pass with requests.

        project (str):  Project name for current project.
    '''
    global _CONFIG_API_KEY, _CONFIG_COOKIES, _CONFIG_PROJECT
    _CONFIG_API_KEY = api_key
    _CONFIG_PROJECT = project
    _CONFIG_COOKIES = {} if cookies is None else cookies


def get_config_variables(api_key, project):
    '''Returns current project config variables unless other is specified.

    Args:
        api_key (str):  Other specified api-key.

        project (str):  Other specified project name.

    Returns:
        tuple: api-key and project name belonging to current project unless other is specified.
    '''
    if api_key is None:
        api_key = _CONFIG_API_KEY
    if project is None:
        project = _CONFIG_PROJECT
    return api_key, project

def get_cookies():
    '''Returns cookies set for the current session.

    Returns:
        dict: Cookies for current session.
    '''
    return _CONFIG_COOKIES

def set_base_url(url):
    '''Sets the base url for requests made from the SDK.

    Args:
        url (str):  URL to set. Set this to None to use default url.
    '''
    global _CONFIG_BASE_URL
    _CONFIG_BASE_URL = url


def get_base_url():
    '''Returns the current base url for requests made from the SDK.

    Returns:
        str: current base url.
    '''
    return BASE_URL if _CONFIG_BASE_URL is None else _CONFIG_BASE_URL


def set_number_of_retries(retries: int):
    '''Sets the number of retries attempted for requests made from the SDK.

    Args:
        retries (int):  Number of retries to attempt.
    '''
    global _CONFIG_RETRIES
    _CONFIG_RETRIES = retries


def get_number_of_retries():
    '''Returns the current number of retries attempted for requests made from the SDK.

    Returns:
        int: current number of retries attempted for requests.
    '''
    return RETRY_LIMIT if _CONFIG_RETRIES is None else _CONFIG_RETRIES



# -*- coding: utf-8 -*-
'''Constants

This module contains constants used in the Cognite Python SDK.

This module is protected and should not used by end-users.

Attributes:
    BASE_URL (str):    Base url for Cognite API. Should have correct version number.
    LIMIT (int):       Limit on how many datapoints should be returned from the API when fetching data using the
                        timeseries module.
    RETRY_LIMIT (int): Number of retries to perform if a request to the API should fail.
'''

# GLOBAL CONSTANTS
BASE_URL = 'https://api.cognitedata.com/api/0.4'
LIMIT = 100000
LIMIT_AGG = 10000
RETRY_LIMIT = 1



# -*- coding: utf-8 -*-
'''Data Objects

This module contains data objects used to represent the data returned from the API. These objects have at least
the following output formats:

    * to_pandas():    Returns pandas dataframe
    * to_ndarray():   Numpy array
    * to_json():      Json format
'''
import abc
import json

import pandas as pd
import six

from cognite import _utils


@six.add_metaclass(abc.ABCMeta)
class CogniteDataObject():
    '''Abstract Cognite Data Object

    This abstract class provides a skeleton for all data objects in this module. All data objects should inherit
    this class.
    '''

    def __init__(self, internal_representation):
        self.internal_representation = internal_representation

    @abc.abstractmethod
    def to_pandas(self):
        '''Returns data as a pandas dataframe'''
        pass

    @abc.abstractmethod
    def to_json(self):
        '''Returns data as a json object'''
        pass

    def to_ndarray(self):
        '''Returns data as a numpy array'''
        return self.to_pandas().values

    def next_cursor(self):
        '''Returns next cursor to use for paging through results'''
        if self.internal_representation.get('data'):
            return self.internal_representation.get('data').get('nextCursor')

    def previous_cursor(self):
        '''Returns previous cursor'''
        if self.internal_representation.get('data'):
            return self.internal_representation.get('data').get('previousCursor')


class RawRowDTO(object):
    """DTO for a row in a raw database.

    The Raw API is a simple key/value-store. Each row in a table in a raw database consists of a
    unique row key and a set of columns.

    Attributes:

        key (str):      Unique key for the row.
        columns (int):  A key/value-map consisting of the values in the row.

    """

    def __init__(self, key, columns):
        self.key = key
        self.columns = columns

    def __repr__(self):
        return json.dumps(self.repr_json())

    def repr_json(self):
        return self.__dict__


class RawObject(CogniteDataObject):
    """Raw Data Object."""

    def to_json(self):
        """Returns data as a json object"""
        return self.internal_representation['data']['items']

    def to_pandas(self):
        """Returns data as a pandas dataframe"""
        return pd.DataFrame(self.internal_representation['data']['items'])


class TagMatchingObject(CogniteDataObject):
    '''Tag Matching Data Object.

    In addition to the standard output formats this data object also has a to_list() method which returns a list of
    names of the tag matches.
    '''

    def to_pandas(self):
        '''Returns data as a pandas dataframe'''
        matches = []
        for tag in self.internal_representation['data']['items']:
            for match in tag['matches']:
                matches.append({
                    'tag': tag['tagId'],
                    'match': match['tagId'],
                    'score': match['score'],
                    'platform': match['platform']
                })
        if matches:
            return pd.DataFrame(matches)[['tag', 'match', 'platform', 'score']]
        return pd.DataFrame()

    def to_json(self):
        '''Returns data as a json object'''
        return self.internal_representation['data']['items']

    def to_list(self, first_matches_only=True):
        '''Returns a list representation of the matches.

        Args:
            first_matches_only (bool):      Boolean determining whether or not to return only the top match for each
                                            tag.

        Returns:
            list: list of matched tags.
        '''
        if self.to_pandas().empty:
            return []
        if first_matches_only:
            return self.to_pandas().sort_values(['score', 'match']).groupby(['tag']).first()['match'].tolist()
        return self.to_pandas().sort_values(['score', 'match'])['match'].tolist()


class DatapointsObject(CogniteDataObject):
    '''Datapoints Object.'''

    def to_json(self):
        '''Returns data as a json object'''
        return self.internal_representation['data']['items'][0]

    def to_pandas(self):
        '''Returns data as a pandas dataframe'''
        return pd.DataFrame(self.internal_representation['data']['items'][0]['datapoints'])


class LatestDatapointObject(CogniteDataObject):
    '''Latest Datapoint Object.'''

    def to_json(self):
        '''Returns data as a json object'''
        return self.internal_representation['data']['items'][0]

    def to_pandas(self):
        '''Returns data as a pandas dataframe'''
        return pd.DataFrame([self.internal_representation['data']['items'][0]])

    def to_ndarray(self):
        '''Returns data as a numpy array'''
        return self.to_pandas().values[0]


class TimeseriesObject(CogniteDataObject):
    '''Timeseries Object'''

    def to_json(self):
        '''Returns data as a json object'''
        return self.internal_representation

    def to_pandas(self):
        '''Returns data as a pandas dataframe'''
        if self.internal_representation[0].get('metadata') is None:
            return pd.DataFrame(self.internal_representation)
        [d.update(d.pop('metadata')) for d in self.internal_representation]
        return pd.DataFrame(self.internal_representation)

class TimeSeriesDTO(object):
    """Data Transfer Object for a timeseries.

    Attributes:
        tag_id (str):       Unique ID of time series.
        is_string (bool):    Whether the time series is string valued or not.
        metadata (dict):    Metadata.
        unit (str):         Physical unit of the time series.
        asset_id (str):     Asset that this time series belongs to.
        description (str):  Description of the time series.
        securityCategories (list(int)): Security categories required in order to access this time series.
        step (bool):        Whether or not the time series is a step series.

    """

    def __init__(self, tag_id, is_string=False, metadata=None, unit=None, asset_id=None, description=None,
                 security_categories=None, step=None):
        self.tagId = tag_id
        self.isString = is_string
        self.metadata = metadata
        self.unit = unit
        self.assetId = asset_id
        self.description = description
        self.security_categories = security_categories
        self.step = step


class DatapointDTO(object):
    '''Data transfer object for datapoints.

    Attributes:
        timestamp (int, datetime): The data timestamp in milliseconds since the epoch (Jan 1, 1970) or as a datetime object.
        value (string):     The data value, Can be string or numeric depending on the metric.
    '''

    def __init__(self, timestamp, value):
        self.timestamp = timestamp if isinstance(timestamp, int) else _utils.datetime_to_ms(timestamp)
        self.value = value
        

class AssetSearchObject(CogniteDataObject):
    '''Assets Search Data Object'''

    def next_cursor(self):
        '''Returns next cursor to use for paging through results'''
        return self.internal_representation['data'].get('nextCursor')

    def previous_cursor(self):
        '''Returns previous cursor'''
        return self.internal_representation['data'].get('previousCursor')

    def to_json(self):
        '''Returns data as a json object'''
        return self.internal_representation['data']['items']

    def to_pandas(self):
        '''Returns data as a pandas dataframe'''
        if len(self.to_json()) > 1:
            return pd.DataFrame(self.internal_representation['data']['items'])
        return pd.DataFrame()


class AssetDTO(object):
    '''Data transfer object for assets.

    Attributes:
        name (str):                 Name of asset. Often referred to as tag.
        parent_id (int):            ID of parent asset, if any.
        description (str):          Description of asset.
        metadata (dict):            Custom , application specific metadata. String key -> String Value.
        ref_id (str):               Reference ID used only in post request to disambiguate references to duplicate
                                    names.
        parent_name (str):          Name of parent, this parent must exist in the same POST request.
        parent_ref_id (list(int)):  Reference ID of parent, to disambiguate if multiple nodes have the same name.
    '''

    def __init__(self, name, parent_id=None, description=None, metadata=None, ref_id=None, parent_name=None,
                 parent_ref_id=None):
        self.name = name
        self.parentId = parent_id
        self.description = description
        self.metadata = metadata
        self.refId = ref_id
        self.parentName = parent_name
        self.parentRefId = parent_ref_id



'''Cognite API SDK

This package has been created to ensure excellent excellent user experience for data scientists using the Cognite
Data Platform (CDP).
'''

# RELEASE NAMING CONVENTIONS
#
# Format:
#
# MAJOR.MINOR[.MICRO]
# The major and minor version numbers should mirror the Cognite API. Micro releases are dedicated to bug fixes,
# improvements, and additions.
#


__all__ = ['assets', 'config', 'data_objects', 'tagmatching', 'timeseries', 'raw', 'preprocessing']
__version__ = '0.4.29'
